/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package cz.fidentis.analyst.procrustes;

import cz.fidentis.analyst.icp.Quaternion;
import javax.vecmath.Vector3d;
import org.ejml.simple.SimpleMatrix;

/**
 * Class used to store information about transformation that needs to be applied
 * to a second face when computing procrustes analysis
 *
 * @author Jakub Kolman
 */
public class ProcrustesTransformation {

    private double scale;
    // used for adjustment to move feature points (0,0,0) so the face is rotated around the centroid point
    private Vector3d centroidAdjustment;
    // used to move face to superimposed poisition over the original human face 
    private Vector3d superImpositionAdjustment;
    private SimpleMatrix rotationMatrix;
    private ProcrustesAnalysisFaceModel faceModel;

    /**
     * constructor
     *
     * @param scale determines by how much should be faced enlarged or minimized
     * @param adjustment determines how should be face moved to make faces
     * superimposed considering feature points
     * @param rotationMatrix determines how to rotate face
     * @param updatedFaceModel holds information about adjusted face
     */
    public ProcrustesTransformation(
            double scale,
            Vector3d centroidAdjustment,
            Vector3d superImpositionAdjustment,
            SimpleMatrix rotationMatrix,
            ProcrustesAnalysisFaceModel updatedFaceModel) {
        this.scale = scale;
        this.centroidAdjustment = centroidAdjustment;                   
        this.superImpositionAdjustment = superImpositionAdjustment;
        this.rotationMatrix = rotationMatrix;
        this.faceModel = updatedFaceModel;
    }

    /**
     * Constructor (for more info take a look on 
     * {@link #ProcrustesTransformation(double, javax.vecmath.Vector3d, org.ejml.simple.SimpleMatrix, cz.fidentis.analyst.procrustes.ProcrustesAnalysisFaceModel) }
     *
     * @param adjustment
     * @param rotationMatrix
     */
    public ProcrustesTransformation(Vector3d centroidAdjustment, Vector3d superImpositionAdjustment, SimpleMatrix rotationMatrix) {
        this.scale = 1;
        this.centroidAdjustment = centroidAdjustment;
        this.superImpositionAdjustment = superImpositionAdjustment;
        this.rotationMatrix = rotationMatrix;
    }

    /**
     * Empty constructor with values that won't transform face. To set values later attribute setters can be used.
     * <pre>
     * rotation matrix is set to identity matrix
     * scale is 1
     * adjustment is none
     * </pre>
     * (for more info take a look on 
     * {@link #ProcrustesTransformation(double, javax.vecmath.Vector3d, org.ejml.simple.SimpleMatrix, cz.fidentis.analyst.procrustes.ProcrustesAnalysisFaceModel) }
     */
    public ProcrustesTransformation() {
        this.scale = 1;
        this.centroidAdjustment = new Vector3d(0,0,0);
        this.superImpositionAdjustment = new Vector3d(0,0,0);
        SimpleMatrix r = new SimpleMatrix(3, 3);    // identity matrix
        r.set(0, 0, 1);
        r.set(0, 1, 0);
        r.set(0, 2, 0);
        r.set(1, 0, 0);
        r.set(1, 1, 1);
        r.set(1, 2, 0);
        r.set(2, 0, 0);
        r.set(2, 1, 0);
        r.set(2, 2, 1);
        this.rotationMatrix = r;
    }

    public double getScale() {
        return this.scale;
    }

    public Vector3d getCentroidAdjustment() {
        return this.centroidAdjustment;
    }

    public Vector3d getSuperImpositionAdjustment() {
        return this.superImpositionAdjustment;
    }

    public SimpleMatrix getRotationMatrix() {
        return this.rotationMatrix;
    }

    public ProcrustesAnalysisFaceModel getFaceModel() {
        return faceModel;
    }

    public void setScale(double scale) {
        this.scale = scale;
    }

    public void setFaceModel(ProcrustesAnalysisFaceModel faceModel) {
        this.faceModel = faceModel;
    }

    /**
     * Calculates quaternion from rotation matrix.
     *
     * @return quaternion with rotation values
     */
    public Quaternion getMatrixAsQuaternion(SimpleMatrix matrix) {
        if (matrix != null) {           
            double w = Math.sqrt(1.0 + matrix.get(0, 0) + matrix.get(1, 1) + matrix.get(2, 2)) / 2.0;
            double w4 = (4.0 * w);
            double x = (matrix.get(2, 1) - matrix.get(1, 2)) / w4;
            double y = (matrix.get(0, 2) - matrix.get(2, 0)) / w4;
            double z = (matrix.get(1, 0) - matrix.get(0, 1)) / w4;
            return new Quaternion(x, y, z, w);
        } else {
            return null;
        }
    }

    public void setCentroidAdjustment(Vector3d centroidAdjustment) {
        this.centroidAdjustment = centroidAdjustment;
    }

    public void setSuperImpositionAdjustment(Vector3d superImpositionAdjustment) {
        this.superImpositionAdjustment = superImpositionAdjustment;
    }

    public void setRotationMatrix(SimpleMatrix rotationMatrix) {
        this.rotationMatrix = rotationMatrix;
    }
    
   public SimpleMatrix getIdentityMatrix() {
        SimpleMatrix r = new SimpleMatrix(3, 3);    // identity matrix
        r.set(0, 0, 1);
        r.set(0, 1, 0);
        r.set(0, 2, 0);
        r.set(1, 0, 0);
        r.set(1, 1, 1);
        r.set(1, 2, 0);
        r.set(2, 0, 0);
        r.set(2, 1, 0);
        r.set(2, 2, 1);
        return r;
   }
}
