/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package cz.fidentis.analyst.visitors.procrustes;

import cz.fidentis.analyst.face.HumanFace;
import cz.fidentis.analyst.mesh.MeshVisitor;
import cz.fidentis.analyst.mesh.core.MeshFacet;
import cz.fidentis.analyst.mesh.core.MeshPoint;
import cz.fidentis.analyst.procrustes.ProcrustesTransformation;
import cz.fidentis.analyst.visitors.mesh.sampling.NoSampling;
import cz.fidentis.analyst.visitors.mesh.sampling.PointSampling;
import cz.fidentis.analyst.procrustes.ProcrustesAnalysis;
import javax.vecmath.Vector3d;
import org.ejml.simple.SimpleMatrix;

/**
 *
 * @author Jakub Kolman
 */
public class ProcrustesVisitor extends MeshVisitor {
    
    private final HumanFace primaryFace;
    private final HumanFace secondaryFace;
    private final PointSampling samplingStrategy;
    private final boolean scale;
    private ProcrustesTransformation transformation = new ProcrustesTransformation();

    /**
     * constructor 
     * 
     * @param primaryFace Face that is used as a from that a secondondary face will be matched to
     * @param secondaryFace Face for which are transformation calculated to be fitted to a primary face
     * @param scale Allows or forbids second face to be scaled
     * @param strategy 
     */
    public ProcrustesVisitor(HumanFace primaryFace, HumanFace secondaryFace, boolean scale, PointSampling strategy) {
        if (primaryFace == null || secondaryFace == null) {
            throw new IllegalArgumentException("Procrustes visitor: one of the faces is null");
        }
        this.primaryFace = primaryFace;
        this.secondaryFace = secondaryFace;
        this.scale = scale;
        this.samplingStrategy = (strategy == null) ? new NoSampling() : strategy;
    }
    
    /**
     * Constructor without PointSampling strategy set. 
     * For more information take a look on {@link #ProcrustesVisitor(cz.fidentis.analyst.face.HumanFace, cz.fidentis.analyst.face.HumanFace, boolean, cz.fidentis.analyst.visitors.mesh.sampling.PointSampling) }
     * 
     * 
     * @param primaryFace
     * @param secondaryFace
     * @param scale 
     */
    public ProcrustesVisitor(HumanFace primaryFace, HumanFace secondaryFace, boolean scale) {
        this(primaryFace, secondaryFace, scale, null);
    }
    
    public ProcrustesTransformation getTransformation() {
        return this.transformation;
    }
    
    @Override
    public void visitMeshFacet(MeshFacet transformedFacet) {
        try {
            ProcrustesAnalysis procrustes = new ProcrustesAnalysis(this.primaryFace, this.secondaryFace, scale);
            this.transformation = procrustes.analyze();
            
//            applyTransformations(this.transformation, transformedFacet);
        } catch (Exception e) {
            System.err.println(e.getMessage());
        }
    }
    
    private void applyTransformations(ProcrustesTransformation transformation, MeshFacet transformedFacet){
        moveFace(transformedFacet, transformation.getCentroidAdjustment());
        scaleFace(transformedFacet, transformation.getScale());
        rotateFace(transformedFacet, transformation.getRotationMatrix());
        moveFace(transformedFacet, transformation.getSuperImpositionAdjustment());
    }
    
    private void moveFace(MeshFacet transformedFacet, Vector3d vector) {
        for (MeshPoint v : transformedFacet.getVertices()) {
            v.getPosition().x = v.getX() + vector.x;
            v.getPosition().y = v.getY() + vector.y;
            v.getPosition().z = v.getZ() + vector.z;
        }
    }

    private void scaleFace(MeshFacet transformedFacet, double scaleFactor) {
        for (MeshPoint point : transformedFacet.getVertices()) {
            point.getPosition().x = point.getX() * scaleFactor;
            point.getPosition().y = point.getY() * scaleFactor;
            point.getPosition().z = point.getZ() * scaleFactor;
        }
    }

    /**
     * Rotates all vertices.
     * <p>
     * For more details check out single vertex rotation {@link #rotateVertex}.
     *
     * @param vertices
     * @param matrix
     */
    private void rotateFace(MeshFacet transformedFacet, SimpleMatrix matrix) {
        for (MeshPoint v: transformedFacet.getVertices()) {
            rotateVertex(v, matrix);
        }
    }
    
    /**
     * Rotates vertex v by simulating matrix multiplication with given matrix
     *
     * @param v
     * @param matrix
     */
    private void rotateVertex(MeshPoint v, SimpleMatrix matrix) {
                    double x = ((v.getX() * matrix.get(0, 0))
                + (v.getY() * matrix.get(1, 0))
                + (v.getZ() * matrix.get(2, 0)));
            double y = ((v.getX() * matrix.get(0, 1))
                    + (v.getY() * matrix.get(1, 1))
                    + (v.getZ() * matrix.get(2, 1)));
            double z = ((v.getX() * matrix.get(0, 2))
                    + (v.getY() * matrix.get(1, 2))
                    + (v.getZ() * matrix.get(2, 2)));
            v.getPosition().x = x;
            v.getPosition().y = y;
            v.getPosition().z = z;
    }
}
